from __future__ import absolute_import, division, print_function

try:
    from .blockwise import blockwise, atop
    from .core import (Array, block, concatenate, stack, from_array, store,
                       map_blocks, to_hdf5, to_npy_stack, from_npy_stack,
                       from_delayed, asarray, asanyarray, PerformanceWarning,
                       broadcast_arrays, broadcast_to, from_zarr, to_zarr)
    from .tiledb_io import from_tiledb, to_tiledb
    from .routines import (take, choose, argwhere, where, coarsen, insert,
                           ravel, roll, unique, squeeze, ptp, diff, ediff1d,
                           gradient, bincount, digitize, histogram, cov, array,
                           dstack, vstack, hstack, compress, extract, round,
                           count_nonzero, flatnonzero, nonzero, unravel_index,
                           around, isin, isnull, notnull, isclose, allclose,
                           corrcoef, swapaxes, tensordot, transpose, dot, vdot,
                           matmul, outer, apply_along_axis, apply_over_axes,
                           result_type, atleast_1d, atleast_2d, atleast_3d,
                           piecewise, flip, flipud, fliplr, einsum, average)
    from .reshape import reshape
    from .ufunc import (add, subtract, multiply, divide, logaddexp, logaddexp2,
            true_divide, floor_divide, negative, power, remainder, mod, conj, exp,
            exp2, log, log2, log10, log1p, expm1, sqrt, square, cbrt, reciprocal,
            sin, cos, tan, arcsin, arccos, arctan, arctan2, hypot, sinh, cosh,
            tanh, arcsinh, arccosh, arctanh, deg2rad, rad2deg, greater,
            greater_equal, less, less_equal, not_equal, equal, maximum,
            bitwise_and, bitwise_or, bitwise_xor, bitwise_not, invert, minimum,
            logical_and, logical_or, logical_xor, logical_not, fmax, fmin,
            isreal, iscomplex, isfinite, isinf, isneginf, isposinf, isnan, signbit,
            copysign, nextafter, spacing, ldexp, fmod, floor, ceil, trunc, degrees,
            radians, rint, fix, angle, real, imag, clip, fabs, sign, absolute,
            i0, sinc, nan_to_num, frexp, modf, divide, frompyfunc, float_power)
    from .reductions import (sum, prod, mean, std, var, any, all, min, max,
                             moment, trace,
                             argmin, argmax,
                             nansum, nanmean, nanstd, nanvar, nanmin,
                             nanmax, nanargmin, nanargmax,
                             cumsum, cumprod,
                             topk, argtopk,
                             nanprod, nancumprod, nancumsum)
    from .percentile import percentile
    from . import ma
    from . import random, linalg, overlap, fft, backends
    from .overlap import map_overlap
    from .wrap import ones, zeros, empty, full
    from .creation import ones_like, zeros_like, empty_like, full_like
    from .rechunk import rechunk
    from ..base import compute
    from .optimization import optimize
    from .creation import (arange, linspace, meshgrid, indices, diag, eye,
                           triu, tril, fromfunction, tile, repeat, pad,
                           diagonal)
    from .gufunc import apply_gufunc, gufunc, as_gufunc
    from .utils import assert_eq

except ImportError as e:
    msg = ("Dask array requirements are not installed.\n\n"
           "Please either conda or pip install as follows:\n\n"
           "  conda install dask                 # either conda install\n"
           "  pip install dask[array] --upgrade  # or pip install")
    raise ImportError(str(e) + '\n\n' + msg)



import pytest
pytest.importorskip('numpy')

import numpy as np

import dask.array as da
from dask.array.utils import assert_eq, same_keys

try:
    import crick
except ImportError:
    crick = None


percentile_methods = pytest.mark.parametrize('method', [
    pytest.param("tdigest", marks=pytest.mark.skipif(not crick, reason="Requires crick")),
    "dask"
])


@percentile_methods
def test_percentile(method):
    d = da.ones((16,), chunks=(4,))
    qs = [0, 50, 100]

    assert_eq(da.percentile(d, qs, method=method),
              np.array([1, 1, 1], dtype=d.dtype))

    x = np.array([0, 0, 5, 5, 5, 5, 20, 20])
    d = da.from_array(x, chunks=(3,))

    result = da.percentile(d, qs, method=method)
    assert_eq(result,
              np.array([0, 5, 20], dtype=result.dtype))

    assert same_keys(da.percentile(d, qs, method=method),
                     da.percentile(d, qs, method=method))
    assert not same_keys(da.percentile(d, qs, method=method),
                         da.percentile(d, [0, 50], method=method))

    if method != 'tdigest':
        x = np.array(['a', 'a', 'd', 'd', 'd', 'e'])
        d = da.from_array(x, chunks=(3,))
        assert_eq(da.percentile(d, [0, 50, 100]),
                  np.array(['a', 'd', 'e'], dtype=x.dtype))


@pytest.mark.skip
def test_percentile_with_categoricals():
    try:
        import pandas as pd
    except ImportError:
        return
    x0 = pd.Categorical(['Alice', 'Bob', 'Charlie', 'Dennis', 'Alice', 'Alice'])
    x1 = pd.Categorical(['Alice', 'Bob', 'Charlie', 'Dennis', 'Alice', 'Alice'])

    dsk = {('x', 0): x0, ('x', 1): x1}

    x = da.Array(dsk, 'x', chunks=((6, 6),))

    p = da.percentile(x, [50])
    assert (p.compute().categories == x0.categories).all()
    assert (p.compute().codes == [0]).all()
    assert same_keys(da.percentile(x, [50]),
                     da.percentile(x, [50]))


@percentile_methods
def test_percentiles_with_empty_arrays(method):
    x = da.ones(10, chunks=((5, 0, 5),))
    assert_eq(da.percentile(x, [10, 50, 90], method=method), np.array([1, 1, 1], dtype=x.dtype))


@percentile_methods
@pytest.mark.parametrize('q', [5, 5.0, np.int64(5), np.float64(5)])
def test_percentiles_with_scaler_percentile(method, q):
    # Regression test to ensure da.percentile works with scalar percentiles
    # See #3020
    d = da.ones((16,), chunks=(4,))
    assert_eq(da.percentile(d, q, method=method), np.array([1], dtype=d.dtype))


@percentile_methods
def test_unknown_chunk_sizes(method):
    x = da.random.random(1000, chunks=(100,))
    x._chunks = ((np.nan,) * 10,)

    result = da.percentile(x, 50, method=method).compute()
    assert 0.1 < result < 0.9

    a, b = da.percentile(x, [40, 60], method=method).compute()
    assert 0.1 < a < 0.9
    assert 0.1 < b < 0.9
    assert a < b
